#include "main.h"

extern unsigned pid;
extern unsigned iterations;
extern unsigned initial_nodes;
extern unsigned backoff;
extern unsigned backoff_base;
extern private_t private;
extern shared_mem_t* smp;

void
init_private()
{
  private.node = 2 + initial_nodes + pid;
  private.value = 1 + initial_nodes + (pid * iterations);

}

void
init_memory()
{
}

unsigned
new_node()
{
  return private.node;
}

void
reclaim(unsigned node)
{
  private.node = node;
}

void
init_queue()
{
  unsigned i;

  /* initialize queue */
  smp->head.sep.ptr = NULL;
  smp->head.sep.count = 0;
  smp->tail.sep.ptr = NULL;
  smp->tail.sep.count = 0;
  
  /* initialize avail list */
  for (i=1; i<MAX_NODES; i++) {
    smp->nodes[i].next.sep.ptr = i+1;
    smp->nodes[i].next.sep.count = 0;
  }
  smp->nodes[MAX_NODES].next.sep.ptr = NULL;
  smp->nodes[MAX_NODES].next.sep.count = 0;
  
  /* initialize queue */
  if (initial_nodes > 0) {
    for (i=1; i<initial_nodes+1; i++) {
      smp->nodes[i].value = i;
      smp->nodes[i].next.sep.ptr = i+1;
    } 
    smp->nodes[initial_nodes].next.sep.ptr = NULL;
    smp->head.sep.ptr = 1;
    smp->tail.sep.ptr = initial_nodes;
  }
}


void
take_snapshot()
{
  pointer_t first_head;
  pointer_t first_tail;

  private.next.con = NULL;
  do {
    first_head.con = smp->head.con;
    do {
      first_tail.con = smp->tail.con;
      if (first_tail.sep.ptr != NULL) {
	private.next.con = smp->nodes[first_tail.sep.ptr].next.con;
      }
      private.tail.con =  smp->tail.con;
    } while (private.tail.con != first_tail.con);
    private.head.con = smp->head.con;
  } while (private.head.con != first_head.con);
}

unsigned
snapshot_state()
{
  unsigned state;

  if (private.tail.sep.ptr == NULL) {
    if (private.head.sep.ptr == NULL) {
      /* Head == Tail == NULL */
      state = 7;
    } else {
      /* Head == DEQ  Tail == NULL */
      state = 6;
    }
  } else {
    if (private.head.sep.ptr == NULL) {
      /* Head == NULL  Tail == ENQ   Next == NULL */
      state = 8;
    } else {
      if (private.next.sep.ptr == NULL) {
	if (private.head.sep.ptr == private.tail.sep.ptr) {
	  if ((private.next.sep.count & MARK_MASK) == ENQUEUE_C) {
	    /* Head == Tail == ENQ   Next == NULL */
	    state = 3;
	  } else {
	    /* Head == Tail == DEQ   Next == NULL */
	    state = 4;
	  }
	} else {
	  /* Head != Tail == ENQ   Next == NULL */ 
	  state = 1;
	}
      } else {
	if (private.head.sep.ptr == private.tail.sep.ptr) {
	  /* Head == Tail == ENQ   Next == ENQ */
	  state = 5;
	} else {
	  /* Head != Tail == ENQ   Next == ENQ */
	  state = 2;
	}
      }
    }
  }
  return state;
}

void
enqueue(unsigned val)
{
  unsigned success;
  unsigned state;
  unsigned node;

  node = new_node();
  smp->nodes[node].value = val;
  smp->nodes[node].next.sep.count += 2;
  smp->nodes[node].next.sep.ptr = NULL;
  smp->nodes[node].next.sep.count &= (~MARK_MASK);
  smp->nodes[node].next.sep.count |= ENQUEUE_C;

  backoff = backoff_base;
  for (success = FALSE; success == FALSE; ) {
    take_snapshot();
    state = snapshot_state();

    switch (state) {
    case 1:
      /* Head != Tail == ENQ   Next == NULL */ 
    case 3:
      /* Head == Tail == ENQ   Next == NULL */
      /* Queue is in correct state, enqueue node */
      success = cas(&smp->nodes[private.tail.sep.ptr].next,
		    private.next.con,
		    MAKE_LONG(node, 
			      CLEAR_MARK(private.next.sep.count)+2+ENQUEUE_C));
      if (success == TRUE) {
	/* Attempt was successful, shift tail to the new node */
	cas(&smp->tail,
	    private.tail.con, 
	    MAKE_LONG(node, private.tail.sep.count+2));
      } else {
	backoff_delay();
      }
      break;
    case 2:
      /* Head != Tail == ENQ   Next == ENQ */
    case 5:
      /* Head == Tail == ENQ   Next == ENQ */
      /* Finish incomplete enqueue operation */
      cas(&smp->tail,
	  private.tail.con, 
	  MAKE_LONG(private.next.sep.ptr, private.tail.sep.count+2));
      backoff_delay();
      break;
    case 4:
      /* Head == Tail == DEQ   Next == NULL */
      /* Cooperate in dequeuing */
      /* shift tail to NULL */
      cas(&smp->tail, 
	  private.tail.con, 
	  MAKE_LONG(NULL, private.tail.sep.count+2));
      /* shift head to NULL */
      cas(&smp->head, 
	  private.head.con, 
	  MAKE_LONG(NULL, private.head.sep.count+2));
      backoff_delay();
      break;
    case 6:
      /* Head == DEQ  Tail == NULL */
      /* Complete dequeuing */
      /* shift head to NULL */
      cas(&smp->head,
	  private.head.con, 
	  MAKE_LONG(NULL, private.head.sep.count+2));
      backoff_delay();
      break;
    case 7:
      /* Head == Tail == NULL */
      /* No elements in queue */
      /* shift tail to node */
      success = cas(&smp->tail, 
	       private.tail.con, 
	       MAKE_LONG(node, private.tail.sep.count+2));
      if (success == TRUE) {
	/* The node has been added, shift head to node */
	cas(&smp->head,
	    private.head.con, 
	    MAKE_LONG(node, private.head.sep.count+2));
      } else {
	backoff_delay();
      }
      break;
    case 8:
      /* Head == NULL  Tail == ENQ   Next == NULL */
      /* Finish incomplete enqueue */
      cas(&smp->head,
	  private.head.con, 
	  MAKE_LONG(private.tail.sep.ptr, private.head.sep.count+2));
      backoff_delay();
      break;
    default:
      assert(0);
    }
  }
}

unsigned
dequeue()
{
  unsigned success;
  unsigned state;
  unsigned val;
  unsigned node;

  backoff = backoff_base;
  for (success = FALSE; success == FALSE; ) {
    take_snapshot();
    state = snapshot_state();
    
    switch (state) {
    case 1:
      /* Head != Tail == ENQ   Next == NULL */ 
    case 2:
      /* Head != Tail == ENQ   Next == ENQ */
      /* Do a sigle step dequeue. i.e. shift head to next object */
      success = cas(&smp->head,
		    private.head.con,
		    MAKE_LONG(smp->nodes[private.head.sep.ptr].next.sep.ptr,
			      private.head.sep.count+2));
      if (success == 1) {
	node = private.head.sep.ptr;
      } else {
	backoff_delay();
      }
      break;
    case 3:
      /* Head == Tail == ENQ   Next == NULL */
      /* Single node in queue, then do a three step dequeue */
      /* Mark the node to be dequeued */
      success = cas(&smp->nodes[private.tail.sep.ptr].next,
		    private.next.con,
		    MAKE_LONG(private.next.sep.ptr,
			      CLEAR_MARK(private.next.sep.count)+2+DEQUEUE_C));
      if (success == 1) {
	node = private.head.sep.ptr;
	/* Attempt was successful, shift tail to NULL */
	cas(&smp->tail,
	    private.tail.con,
	    MAKE_LONG(NULL, private.tail.sep.count+2));
	/* Shift head to NULL */
	cas(&smp->head,
	    private.head.con, 
	    MAKE_LONG(NULL, private.head.sep.count+2));
      } else {
	backoff_delay();
      }
      break;
    case 4:
      /* Head == Tail == DEQ   Next == NULL */
      /* Cooperate in dequeuing */
      /* shift tail to NULL */
      cas(&smp->tail,
	  private.tail.con,
	  MAKE_LONG(NULL, private.tail.sep.count+2));
      /* shift head to NULL */
      cas(&smp->head,
	  private.head.con,
	  MAKE_LONG(NULL, private.head.sep.count+2));
      backoff_delay();
      break;
    case 5:
      /* Head == Tail == ENQ   Next == ENQ */
      /* Complete incomplete enqueue opertation */
      cas(&smp->tail, 
	  private.tail.con,
	  /* MAKE_LONG(smp->nodes[private.tail.sep.ptr].next.sep.ptr,  */
	  MAKE_LONG(private.next.sep.ptr, 
		    private.tail.sep.count+2));
      backoff_delay();
      break;
    case 6:
      /* Head == DEQ  Tail == NULL */
      /* Complete dequeuing */
      /* shift head to NULL */
      cas(&smp->head,
	  private.head.con, 
	  MAKE_LONG(NULL, private.head.sep.count+2));
      backoff_delay();
      break;
    case 7:
      /* Head == Tail == NULL */
      /* No elements in queue */
      node = NULL;
      success = TRUE;
      break;
    case 8:
      /* Head == NULL  Tail == ENQ   Next == NULL */
      /* Finish incomplete enqueue */
      cas(&smp->head,
	  private.head.con,
	  MAKE_LONG(private.tail.sep.ptr, private.head.sep.count+2));
      backoff_delay();
      break;
    default:
      assert(0);
    }
  }
  /* Delink dequeued node */
  if (node != NULL ) {
    smp->nodes[node].next.sep.ptr = NULL;
    val = smp->nodes[node].value;
  }
  reclaim(node);
  return val;
}
